﻿using System;
using System.Collections.Generic;
using System.Dynamic;
using System.IO;
using System.Linq;
using System.Linq.Expressions;
using System.Text;
using System.Threading.Tasks;
using System.Xml.Linq;

namespace Chapter4
{
    class DynamicPropertyBag : DynamicObject
    {
        private Dictionary<string, object> storage =
            new Dictionary<string, object>();

        public override bool TryGetMember(GetMemberBinder binder,
            out object result)
        {
            if (storage.ContainsKey(binder.Name))
            {
                result = storage[binder.Name];
                return true;
            }
            result = null;
            return false;
        }

        public override bool TrySetMember(SetMemberBinder binder,
            object value)
        {
            string key = binder.Name;
            if (storage.ContainsKey(key))
                storage[key] = value;
            else
                storage.Add(key, value);
            return true;
        }

        public override string ToString()
        {
            StringWriter message = new StringWriter();
            foreach (var item in storage)
                message.WriteLine($"{item.Key}:\t{item.Value}");
            return message.ToString();
        }
    }

    public class DynamicXElement : DynamicObject
    {
        private readonly XElement xmlSource;

        public DynamicXElement(XElement source)
        {
            xmlSource = source;
        }

        public override bool TryGetMember(GetMemberBinder binder,
            out object result)
        {
            result = new DynamicXElement(null);
            if (binder.Name == "Value")
            {
                result = (xmlSource != null) ? xmlSource.Value : "";
                return true;
            }
            if (xmlSource != null)
                result = new DynamicXElement(xmlSource.Element(XName.Get(binder.Name)));
            return true;
        }

        public override bool TryGetIndex(GetIndexBinder binder, object[] indexes, out object result)
        {
            result = null;
            // This only supports [string, int] indexers
            if (indexes.Length != 2)
                return false;
            if (!(indexes[0] is string))
                return false;
            if (!(indexes[1] is int))
                return false;

            var allNodes = xmlSource.Elements(indexes[0].ToString());
            int index = (int)indexes[1];
            if (index < allNodes.Count())
                result = new DynamicXElement(allNodes.ElementAt(index));
            else
                result = new DynamicXElement(null);
            return true;
        }

        public override string ToString() =>
            xmlSource?.ToString() ?? string.Empty;
    }

    class DynamicDictionary2 : IDynamicMetaObjectProvider
    {
        DynamicMetaObject IDynamicMetaObjectProvider.GetMetaObject(
            System.Linq.Expressions.Expression parameter)
        {
            return new DynamicDictionaryMetaObject(parameter, this);
        }

        private Dictionary<string, object> storage = new Dictionary<string, object>();

        public object SetDictionaryEntry(string key, object value)
        {
            if (storage.ContainsKey(key))
                storage[key] = value;
            else
                storage.Add(key, value);
            return value;
        }

        public object GetDictionaryEntry(string key)
        {
            object result = null;
            if (storage.ContainsKey(key))
            {
                result = storage[key];
            }
            return result;
        }

        private class DynamicDictionaryMetaObject : DynamicMetaObject
        {
            internal DynamicDictionaryMetaObject(
                System.Linq.Expressions.Expression parameter,
                DynamicDictionary2 value)
                : base(parameter, BindingRestrictions.Empty, value)
            {
            }

            public override DynamicMetaObject BindSetMember(SetMemberBinder binder,
                DynamicMetaObject value)
            {
                // Method to call in the containing class:
                string methodName = "SetDictionaryEntry";

                // setup the binding restrictions.
                BindingRestrictions restrictions =
                    BindingRestrictions.GetTypeRestriction(Expression, LimitType);

                // setup the parameters:
                Expression[] args = new Expression[2];
                // First parameter is the name of the property to Set
                args[0] = Expression.Constant(binder.Name);
                // Second parameter is the value
                args[1] = Expression.Convert(value.Expression, typeof(object));

                // Setup the 'this' reference
                Expression self = Expression.Convert(Expression, LimitType);

                // Setup the method call expression
                Expression methodCall = Expression.Call(self,
                        typeof(DynamicDictionary2).GetMethod(methodName),
                        args);

                // Create a meta object to invoke Set later:
                DynamicMetaObject setDictionaryEntry = new DynamicMetaObject(
                    methodCall,
                    restrictions);
                // return that dynamic object
                return setDictionaryEntry;
            }

            public override DynamicMetaObject BindGetMember(GetMemberBinder binder)
            {
                // Method call in the containing class:
                string methodName = "GetDictionaryEntry";

                // One parameter
                Expression[] parameters = new Expression[]
                {
                Expression.Constant(binder.Name)
                };

                DynamicMetaObject getDictionaryEntry = new DynamicMetaObject(
                    Expression.Call(
                        Expression.Convert(Expression, LimitType),
                        typeof(DynamicDictionary2).GetMethod(methodName),
                        parameters),
                    BindingRestrictions.GetTypeRestriction(Expression, LimitType));
                return getDictionaryEntry;
            }
        }

        public override string ToString()
        {
            StringWriter message = new StringWriter();
            foreach (var item in storage)
                message.WriteLine($"{item.Key}:\t{item.Value}");
            return message.ToString();
        }
    }

    public class Item42
    {
        public static void Snippet01()
        {
            dynamic dynamicProperties = new DynamicPropertyBag();

            try
            {
                Console.WriteLine(dynamicProperties.Marker);
            }
            catch (Microsoft.CSharp.RuntimeBinder.RuntimeBinderException)
            {
                Console.WriteLine("There are no properties");
            }

            dynamicProperties.Date = DateTime.Now;
            dynamicProperties.Name = "Bill Wagner";
            dynamicProperties.Title = "Effective C#";
            dynamicProperties.Content = "Building a dynamic dictionary";
        }

        public static void Snippet02()
        {
            // Create an XElement document containing
            // solar system data:
            var xml = createXML();

            Console.WriteLine(xml);

            dynamic dynamicXML = new DynamicXElement(xml);

            // old way:
            var firstPlanet = xml.Element("Planet");
            Console.WriteLine(firstPlanet);
            // new way:
            dynamic test2 = dynamicXML.Planet; // returns the first planet.
            Console.WriteLine(test2);

            // gets the third planet (Earth)
            dynamic test3 = dynamicXML["Planet", 2];

            dynamic earthMoon = dynamicXML["Planet", 2]["Moons", 0].Moon;

            dynamic test6 = dynamicXML["Planet", 2]
                ["Moons", 3].Moon; // earth doesn't have 4 moons
            dynamic fail = dynamicXML.NotAppearingInThisFile;
            dynamic fail2 = dynamicXML.Not.Appearing.In.This.File;

        }

        public static void Snippet03()
        {
            dynamic propertyBag2 = new DynamicDictionary2();

            DateTime current = propertyBag2.Date = DateTime.Now;

            propertyBag2.MagicNumber = GetMagicNumber();
        }

        private static int GetMagicNumber() => 42;

        private static XElement createXML()
        {
            var xml = new XElement("Planets",
                new XElement("Planet",
                    new XElement("Name", "Mercury")),
                new XElement("Planet",
                    new XElement("Name", "Venus")),
                new XElement("Planet",
                    new XElement("Name", "Earth"),
                    new XElement("Moons",
                        new XElement("Moon", "Moon"))),
                new XElement("Planet",
                    new XElement("Name", "Mars"),
                    new XElement("Moons",
                        new XElement("Moon", "Phobos"),
                        new XElement("Moon", "Deimos"))),
                new XElement("Planet",
                    new XElement("Name", "Jupiter"),
                    new XElement("Moons",
                        new XElement("Moon", "Io"),
                        new XElement("Moon", "Europa"),
                        new XElement("Moon", "Ganymede"),
                        new XElement("Moon", "Callisto"),
                        new XElement("Moon", "Leda"),
                        new XElement("Moon", "Himalia"),
                        new XElement("Moon", "Lysithea"),
                        new XElement("Moon", "Elara"),
                        new XElement("Moon", "Ananke"),
                        new XElement("Moon", "Carme"),
                        new XElement("Moon", "Pasiphae"),
                        new XElement("Moon", "Sinope"),
                        new XElement("Moon", "Metis"),
                        new XElement("Moon", "Adrastea"),
                        new XElement("Moon", "Amalthea"),
                        new XElement("Moon", "Thebe"))),
                new XElement("Planet",
                    new XElement("Name", "Saturn"),
                    new XElement("Moons",
                        new XElement("Moon", "Mimas"),
                        new XElement("Moon", "Enceladus"),
                        new XElement("Moon", "Tethys"),
                        new XElement("Moon", "Dione"),
                        new XElement("Moon", "Rhea"),
                        new XElement("Moon", "Titan"),
                        new XElement("Moon", "Iapetus"))),
                new XElement("Planet",
                    new XElement("Name", "Uranus"),
                    new XElement("Moons",
                        new XElement("Moon", "Miranda"),
                        new XElement("Moon", "Ariel"),
                        new XElement("Moon", "Umbriel"),
                        new XElement("Moon", "Titania"),
                        new XElement("Moon", "Oberon"))),
                new XElement("Planet",
                    new XElement("Name", "Neptune"))
                        );
            return xml;
        }

    }
}
